from dios.dios import *
from abc import abstractmethod


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data


# #############################################################################


class _LocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        if isinstance(rowkey, tuple):
            raise KeyError(f"{key}. tuples are not allowed.")
        if is_dios_like(rowkey) or is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        data = self._dios[colkey]

        # .loc[:, any] - simple low-cost optimization
        if isinstance(rowkey, slice) and rowkey == slice(None):
            new = data.copy()

        # .loc[any, scalar]
        elif isinstance(data, pd.Series):
            new = data.loc[rowkey]

        # .loc[any, non-scalar]
        elif isinstance(data, self._dios.__class__):

            # .loc[scalar, non-scalar]
            if is_hashable(rowkey):
                # we do not override data directly to may get
                # a better fitting series dtype
                new = pd.Series(index=type(data.columns)([]))
                for k in data.index:
                    s = data._data.at[k]
                    new.at[k] = s.loc[rowkey]

            # .loc[non-scalar, non-scalar]
            else:
                new = self._dios.copy_empty()
                for k in data.columns:
                    s = data._data.at[k]
                    new._data.at[k] = s.loc[rowkey]
        else:
            raise AssertionError(f"getitem returned data of type {type(data)}")

        return new

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")

            rowkey, colkey = key

            if isinstance(rowkey, tuple):
                raise KeyError(f"{key}. tuples are not allowed.")

            # .loc[dont-care, new-scalar] = val
            # if a unknown colkey was given, we insert it and ignore rowkey
            if is_hashable(colkey) and colkey not in self._dios.columns:
                self._dios._insert(colkey, value)
                return
        else:
            rowkey, colkey = key, slice(None)

        # get .loc[any,any] - we use key(!) here
        data = self.__getitem__(key)

        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".loc[] cannot be used to set multi-dimensional values, use .aloc[] instead.")

        # .loc[scalar, any]
        if is_hashable(rowkey):

            # .loc[scalar, scalar]
            if is_hashable(colkey):
                s = self._data.at[colkey]
                s.at[rowkey] = value
                self._data.at[colkey] = s

            # .loc[scalar, non-scalar] - column-labeled series
            else:
                data.loc[:] = value
                for k in data.index:
                    s = self._data.at[k]
                    s.at[rowkey] = data.at[k]
                    self._data.at[k] = s

        # .loc[non-scalar, scalar] - single row-labeled series
        elif is_hashable(colkey):
            data.loc[rowkey] = value
            self._data.at[colkey] = data

        # .loc[non-scalar, non-scalar]
        elif isinstance(data, self._dios.__class__):
            for k in data.columns:
                s = data._data.at[k]
                s.loc[rowkey] = value
                self._data.at[k] = s

        else:
            raise AssertionError(f"getitem returned data of type {type(data)}")


# #############################################################################


class _iLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        if isinstance(rowkey, tuple):
            raise KeyError(f"{key}. tuples are not allowed.")
        if is_dios_like(rowkey) or is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        if is_bool_indexer(colkey):
            data = self._dios[colkey]._data
        else:
            data = self._data.iloc[colkey]

        # in any case data is a series now,
        # either a column-indexed series of series,
        # or a simple single row-indexed series (of values)
        if isinstance(data, pd.Series):

            # .iloc[any, int]   - got a single row indexed series
            if is_integer(colkey):
                new = data.iloc[rowkey]

            # .loc[int, any]
            if is_integer(colkey):
                # we do not override data directly to may get
                # a better fitting series dtype
                new = pd.Series(index=type(data.index)([]))
                for k in data.index:
                    s = data.at[k]
                    new.at[k] = s.iloc[rowkey]

            # .iloc[:, any] - simple low-cost optimization
            elif isinstance(rowkey, slice) and rowkey == slice(None):
                new = self._dios.copy_empty()
                new._data = data.copy()

            # .loc[non-int, non-int]
            else:
                new = self._dios.copy_empty()
                for k in data.index:
                    new._data.at[k] = data.at[k].iloc[rowkey]
        else:
            raise AssertionError(f"getitem returned data of type {type(data)}")

        return new

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        # get .iloc[any,any] - we use key(!) here
        data = self.__getitem__(key)

        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".loc[] cannot be used to set multi-dimensional values, use .aloc[] instead.")

        # .iloc[scalar, any]
        if is_integer(rowkey):

            # .iloc[scalar, scalar]
            if is_integer(colkey):
                s = self._data.iat[colkey]
                s.iat[rowkey] = value
                self._data.iat[colkey] = s

            # .iloc[scalar, non-scalar] - column-labeled series
            else:
                data.iloc[:] = value
                for k in data.index:
                    s = self._data.at[k]
                    s.iat[rowkey] = data.at[k]
                    self._data.at[k] = s

        # .iloc[non-scalar, scalar] - single row-labeled series
        elif is_integer(colkey):
            data.iloc[rowkey] = value
            self._data.iat[colkey] = data

        # .iloc[non-scalar, non-scalar]
        elif isinstance(data, self._dios.__class__):
            for k in data.columns:
                s = data._data.at[k]
                s.iloc[rowkey] = value
                self._data.at[k] = s

        else:
            raise AssertionError(f"getitem returned data of type {type(data)}")



# #############################################################################


class _aLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        # fallback to loc
        if not is_dios_like(value) and not is_nested_list_like(value):
            self._dios.loc[key] = value

    def __getitem__(self, item):
        return item


# #############################################################################


class _AtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and is_hashable(key[0]) and is_hashable(key[1])):
            raise KeyError(f"{key}. `.at` takes exactly one scalar row-key "
                           "and one scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.at[key[1]].at[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".at[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        s = self._data.at[key[1]]
        s.at[key[0]] = value
        self._data.at[key[1]] = s


# #############################################################################


class _iAtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and is_integer(key[0]) and is_integer(key[1])):
            raise KeyError(f"{key} `.iat` takes exactly one integer positional "
                           f"row-key and one integer positional scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.iat[key[1]].iat[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".iat[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        s = self._data.iat[key[1]]
        s.iat[key[0]] = value
        self._data.iat[key[1]] = s


# #############################################################################


def _unpack_value(keys, ix, val):
    """Return a generator that yield (column key, corresponding value, value-align(bool) )
    for all columns.
    This is analogous to DictOfSeries._unpack_value, but with some modifications."""

    # prepare value
    val = list(val) if is_iterator(val) else val
    val = val.squeeze(axis=1) if is_dios_like(val) else val
    dioslike, nlistlike = is_dios_like(val), is_nested_list_like(val)

    # check value
    if nlistlike and len(val) != len(keys):
        raise ValueError(f"could not broadcast input array with length {len(val)}"
                         f" into dios of length {len(keys)}")
    if dioslike:
        keys = val.columns.intersection(keys)

    for i, k in enumerate(keys):
        if dioslike:
            yield k, ix, val[k]
        elif nlistlike:
            yield k, ix, val[i]
        else:
            yield k, ix, val


def maby_set_series_name(maybe_ser, name):
    if isinstance(maybe_ser, pd.Series):
        maybe_ser.name = name
